// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of src
//
// src is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// src is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <sstream>

#include "Bp2Fits.h"

#include <funtools.h>

#include "FunParamWrap.h"


static std::string makeKeyWord( const string& keyword, bool _verbose );
static std::string makeKeyValue( const string& keyword, const string& value, size_t& offset,
                                        size_t maxsize, bool contchar, bool _verbose, bool _longstrn);

void Bp2Fits::writeHDU_longstrn( 
  bool primary
  ) {

  TraceFct tf( "Bp2Fits::writeMyHDU" );


  if ( _longstrn ) {

#define myFUN_PRIMARY(fun) ((primary) ? FUN_PRIMARY( fun ) : fun)

    ParamPut( myFUN_PRIMARY(_fun), "LONGSTRN", 0, "OGIP 1.0", "The OGIP long string convention may be used.", 1 );
    ParamPut( myFUN_PRIMARY(_fun), "COMMENT",  0, "This FITS file may contain long string keyword values that are", "", 1 );
    ParamPut( myFUN_PRIMARY(_fun), "COMMENT",  0, "continued over multiple keywords.  This convention uses the  '&'", "", 1 );
    ParamPut( myFUN_PRIMARY(_fun), "COMMENT",  0, "character at the end of a string which is then continued", "", 1 );
    ParamPut( myFUN_PRIMARY(_fun), "COMMENT",  0, "on subsequent keywords whose name = 'CONTINUE'.", "", 1 );

  }

}

void Bp2Fits::writeHDU_colinfo( 
  void
  ) {

  TraceFct tf( "Bp2Fits::writeReservedHDU" );

  vector<ColumnInfo*>::iterator cip;
  size_t idx;
  for ( idx = 1, cip = _ci.begin(); cip != _ci.end(); cip++, idx++ )
  {
    ColumnInfo* ci = *cip;

    char type = ci->dtype;

    if ( ci->unit.size() ) 
      ParamPut( _fun, "TUNIT", idx, ci->unit, "", 1 );

    // if this is an unsigned integer column, scale and zero are
    // written to the header by convertDpktf2Col
    if ( ci->cxxdt != DT_uint )
    {
      ParamPutType( _fun, type, "TSCAL", idx, ci->scale, "", 1 );
      ParamPutType( _fun, type, "TZERO", idx, ci->zero, "", 1 );
    }
    ParamPut( _fun, "TDISP", idx, ci->disp, "", 1 );
    ParamPut( _fun, "TDIM",  idx, ci->dim,  "", 1 );

    if ( ci->null.size() ) ParamPutType( _fun, type, "TNULL", idx, ci->null, "", 1 );

    // Add in reserved keywords...
    if ( _reserved ) {

      if ( ci->lmin.size() ) 
	ParamPutType( _fun, type, "TLMIN", idx, ci->lmin, "", 1 );
      else if ( _doublepass )
	ParamPutType( _fun, type, "TLMIN", idx, "0", "", 1 );

      if ( ci->lmax.size() ) 
	ParamPutType( _fun, type, "TLMAX", idx, ci->lmax, "", 1 );
      else if ( _doublepass )
	ParamPutType( _fun, type, "TLMAX", idx, "0", "", 1 );

      // write out data min and max keywords.  if we're doing two
      // passes and it hasn't been specified, set them equal to
      // "innocuous" values

      if ( ci->dmin.size() ) 
	ParamPutType( _fun, type, "TDMIN", idx, ci->dmin, "", 1 );
      else if ( _doublepass )
	ParamPutType( _fun, type, "TDMIN", idx, "0", "", 1 );

      if ( ci->dmax.size() ) 
	ParamPutType( _fun, type, "TDMAX", idx, ci->dmax, "", 1 );
      else if ( _doublepass )
	ParamPutType( _fun, type, "TDMAX", idx, "0", "", 1 );
    }
  }

}

void Bp2Fits::writeHDU_comments( 
  bool primary
  ) {

  TraceFct tf( "Bp2Fits::writeHDU_comments" );

  Fun fun = NULL;

  vector<Comment>::iterator comment;

  // Add in comments from translation file...
  for ( comment = _comments.begin() ;
	comment != _comments.end() ;
	comment ++ ) {

      string &value = comment->value;
      string &keyword = comment->keyword;

    if ( value.npos != value.find_first_not_of( " \t" ) ) {
      // Handle %date, %time, and %user in a special manner...

      // %date and %time now do both date+time and are the same.  seems FITS is going
      // towards this format in general.
      if ( value == "%date" || value == "%time") {
	char val[20];
	time_t now;
	now = time( NULL );
	
	strftime( val, 20, "%Y-%m-%dT%H:%M:%S", gmtime( &now ) );
	val[19] = '\0';
	value = val;
	
      } else if ( value == "%user" ) {
	char* val = getlogin( );
	if ( ! val ) {
	  passwd* pws = getpwuid( getuid( ) );
	  value = pws->pw_name;
	  
	} else 
	  value = val; 
	
      }
      
      size_t offset = ((keyword.substr( 0, 9 ) == "PRIMARY::") ? (9) : (0));
      
      // Some logic to put some cards into the PHDU and other into the
      // CHDU...
      if ( (offset && primary) || (! offset && ! primary) ) {
        auto name = makeKeyWord( keyword.substr( offset ), _verbose );
	offset = 0;
	do {
	  fun = ( (primary) ? (FUN_PRIMARY( _fun )) : (_fun) );
	  if ( name == "COMMENT" || name == "HISTORY" ) {
            auto val = makeKeyValue( keyword, value, offset, 70, false, _verbose, _longstrn );
	    ParamPut( fun, name, 0, val, "", 1 );
	  } else {
            auto val = makeKeyValue( keyword, value, offset, 68, true, _verbose, _longstrn );
	    ParamPut( fun, name, 0, val, "", 1 );
	    name = makeKeyWord( "CONTINUE", _verbose );
	  }
	} while ( _longstrn && offset < value.size( ) );
      }
    }
  }
}

void Bp2Fits::writeHDU_hdrf( 
  void
  ) {

  TraceFct tf( "Bp2Fits::writeHDU_hdrf" );

  const char* keyword = NULL;
  void* _last = NULL;

  size_t idx;
  while ( bpipe_hdrf_next( _bpipe, &keyword, &idx, &_last ) ) {
    BPDataType type  = bpipe_hdrf_type( _bpipe, keyword, idx );
    bpcheck( );
    BPMatrix* matrix = bpipe_hdrf_matrix ( _bpipe, keyword, idx );
    bpcheck( );
    void* data       = bpipe_hdrf_data( _bpipe, keyword, idx );
    bpcheck( );
      
    size_t offset = 0;

    auto name = makeKeyWord( keyword, _verbose );
    switch ( type ) {
    case BPDType_char:
      do {
	if ( name == "COMMENT"  || name == "HISTORY" ) {
          auto val = makeKeyValue( keyword, (char *)data, offset, 70, false, _verbose, _longstrn );
	  ParamPut( _fun, name, 0, val, "", 1 );
	} else {
          auto val = makeKeyValue( keyword, (char *)data, offset, 68, true, _verbose, _longstrn );
	  ParamPut( _fun, name, 0, val, "", 1 );
	  name = makeKeyWord( "CONTINUE", _verbose );
	}
      } while ( _longstrn && offset < matrix->ne );
      break;  

    case BPDType_double:
      ParamPut( _fun, name, 0, *(double *)data, "", 1 );
      break;  
 
    case BPDType_int:
      ParamPut( _fun, name, 0, *(int *)data, "", 1 );
      break;  
 
    case BPDType_uint:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type unsigned int as int\n" );
      ParamPut( _fun, name, 0, *(unsigned int *)data, "", 1 );
      break;  
 
    case BPDType_DVector2:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type DVector2 as double\n" );
      ParamPut( _fun, name, 0, *(double *)data, "", 1 );
      break;  
 
    case BPDType_DVector3:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type DVector3 as double\n" );
      ParamPut( _fun, name, 0, *(double *)data, "", 1 );
      break;  
 
    case BPDType_IVector2:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type IVector2 as int\n" );
      ParamPut( _fun, name, 0, *(int *)data, "", 1 );
      break;  
 
    case BPDType_IVector3:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type IVector3 as int\n" );
      ParamPut( _fun, name, 0, *(int *)data, "", 1 );
      break;  
 
    case BPDType_UIVector2:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type UIVector2 as int\n" );
      ParamPut( _fun, name, 0, *(unsigned int *)data, "", 1 );
      break;  
 
    case BPDType_UIVector3:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type UIVector3 as int\n" );
      ParamPut( _fun, name, 0, *(unsigned int *)data, "", 1 );
      break;  
 
    case BPDType_DComplex:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type DComplex as double\n" );
      ParamPut( _fun, name, 0, *(double *)data, "", 1 );
      break;  
 
    case BPDType_DCVector2:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type DCVector2 as double\n" );
      ParamPut( _fun, name, 0, *(double *)data, "", 1 );
      break;  
 
    case BPDType_DCVector3:
      if ( _verbose )
	TraceFct::message( string( "WARNING: Representing hdrf `" ) + keyword + "' of type DCVector3 as double\n" );
      ParamPut( _fun, name, 0, *(double *)data, "", 1 );
      break;

    default: 
      throw( Bp2FitsErr( string( "Unhandled datatype `" ) + bpipe_datatype_name( type ) + "'" ) );
    }
    bpipe_matrix_delete( matrix );

  }
}

void Bp2Fits::rewriteHDU( 
  void
  ) {

  TraceFct tf( "Bp2Fits::rewriteTable" );

  // If necessary, rewrite the HDU...
  if ( _reserved && _doublepass ) {

    // Card updates go here...
    vector<ColumnInfo*>::iterator cip;
    size_t idx;
    for ( idx = 1, cip = _ci.begin(); cip != _ci.end(); cip++, idx++ )
    {

      ColumnInfo* ci = *cip;

      char* name = NULL;
      int type;
      int funret = FunColumnLookup( _fun, NULL, static_cast<int>(idx-1), &name, &type, NULL, NULL, NULL, NULL );

      if ( ! funret )
      {
	stringstream ss;
	ss << "internal error: unable to lookup column " << idx;
	throw( Bp2FitsErr( ss.str() ) );
      }

      try {

	// use override for tlmin and tlmax, if specified. otherwise
	// just copy TDMIN and TDMAX; this is done below on a type
	// specific basis.
	if ( ci->lmin.size() )
	  ParamPutType( _fun, ci->dtype, "TLMIN", idx, ci->lmin, "", 1 );
	if ( ci->lmax.size() )
	  ParamPutType( _fun, ci->dtype, "TLMAX", idx, ci->lmax, "", 1 );

	switch ( ci->dtype ) {

	case 'A':
          ParamPut( _fun, "TDMIN", idx, reinterpret_cast<const char*>(ci->min), "", 0 );
	  ParamPut( _fun, "TDMAX", idx, reinterpret_cast<const char*>(ci->max), "", 0 );

	  if ( ! ci->lmin.size() )
              ParamPut( _fun, "TLMIN", idx, reinterpret_cast<const char*>(ci->min), "", 0 );
	  if ( ! ci->lmax.size() )
              ParamPut( _fun, "TLMAX", idx, reinterpret_cast<const char*>(ci->max), "", 0 );
	  break;

	case 'D':
	  ParamPut( _fun, "TDMIN", idx, *(double *)ci->min, "", 0 );
	  ParamPut( _fun, "TDMAX", idx, *(double *)ci->max, "", 0 );

	  if ( ! ci->lmin.size() )
	    ParamPut( _fun, "TLMIN", idx, *(double *)ci->min, "", 0 );
	  if ( ! ci->lmax.size() )
	    ParamPut( _fun, "TLMAX", idx, *(double *)ci->max, "", 0 );
	  break;

	case 'J':
	  switch ( ci->cxxdt )
	  {
	  case DT_int:
	    { 
	      int min = *(int *)ci->min;
	      int max = *(int *)ci->max;
	      ParamPut( _fun, "TDMIN", idx, min, "", 0 );
	      ParamPut( _fun, "TDMAX", idx, max, "", 0 );

	      if ( ! ci->lmin.size() )
		ParamPut( _fun, "TLMIN", idx, min, "", 0 );
	      if ( ! ci->lmax.size() )
		ParamPut( _fun, "TLMAX", idx, max, "", 0 );
	    }
	    break;

	    // we need to worry about the fact that unsigned ints are
	    // too big for ints; use double instead
	  case DT_uint:
	    { 
	      double min = *(int *)ci->min;
	      double max = *(int *)ci->max;
	      ParamPut( _fun, "TDMIN", idx, min, "", 0 );
	      ParamPut( _fun, "TDMAX", idx, max, "", 0 );

	      if ( ! ci->lmin.size() )
		ParamPut( _fun, "TLMIN", idx, min, "", 0 );
	      if ( ! ci->lmax.size() )
		ParamPut( _fun, "TLMAX", idx, max, "", 0 );
	    }
	    break;

	  default:
	    throw( Exception( "internal error; J FITS type, but not int or uint C++ type" ) );
	  }
	  break;

	case 'V':
	  ParamPut( _fun, "TDMIN", idx, *(unsigned int *)ci->min, "", 0 );
	  ParamPut( _fun, "TDMAX", idx, *(unsigned int *)ci->max, "", 0 );

	  if ( ! ci->lmin.size() )
	    ParamPut( _fun, "TLMIN", idx, *(unsigned int *)ci->min, "", 0 );
	  if ( ! ci->lmin.size() )
	    ParamPut( _fun, "TLMAX", idx, *(unsigned int *)ci->max, "", 0 );
	  break;

	default:
	  throw( Bp2FitsErr( string( "Unhandled FITS datatype `" ) + ci->dtype + "'" ) );

	}
      }
      catch ( Exception& e )
      {
	e.set_message( (string) "error with column " + name + ": " );
	throw( e );
      }

    }
  }
}


static 
char* xdupstr( string str )
{
  char *nstr = new char[str.size()+1];
  strcpy( nstr, str.c_str() );
  return nstr;
}

// give Funtools the column specifications
void Bp2Fits::addColumns(
  void
)
{
  TraceFct tf( "Bp2Fits::addColumns" );

  if ( _nrows || 0 == _ci.size() )
    return;

  // we need to convert from our structure based data to array based
  // for Funtools

  char** type = new char*[_ncols];
  char** form = new char*[_ncols];
  char** mode = new char*[_ncols];
  int* offset = new int[_ncols];

  vector<ColumnInfo*>::iterator cip;
  size_t idx;
  for ( idx = 0, cip = _ci.begin(); cip != _ci.end(); cip++, idx++ )
  {
    ColumnInfo* ci = *cip;

    type[idx] = xdupstr( ci->type );
    form[idx] = xdupstr( ci->form );
    mode[idx] = xdupstr( ci->mode );
    offset[idx] = ci->offset;
  }

  if ( 0 == FunColumnSelectArr( _fun, _ncols * sizeof( void* ), NULL, 
                                type, form, mode, offset, static_cast<int>(_ncols) ) )
    throw( Exception("Error selecting columns via FunColumnSelectArr") );


  // free up 
  for ( idx = 0 ; idx < _ci.size() ; idx++ )
  {
    delete [] type[idx];
    delete [] form[idx];
    delete [] mode[idx];
  }

  delete [] type;
  delete [] form;
  delete [] mode;
  delete [] offset;
}



static std::string
makeKeyWord( const string& keyword, const bool _verbose ) {

    auto length = keyword.size();
    if ( _verbose && 8 < length )  {
        TraceFct tf( "makeKeyWord" );
        TraceFct::message( string( "WARNING: Truncating BPipe hdrf name `" ) + keyword + "'\n" );
    }

    
    std::string fitskeyword = keyword;

  for ( size_t idx = 0; idx < length; idx++ ) {
    if( 97 <= fitskeyword[idx] && fitskeyword[idx] <= 122 ) {
      fitskeyword[idx] -= 32;

    } else if ( 
      45 != fitskeyword[idx] && 
      95 != fitskeyword[idx] && 
      ( 65 > fitskeyword[idx] || fitskeyword[idx] > 90 ) &&
      ( 48 > fitskeyword[idx] || fitskeyword[idx] > 57 ) 
      ) {
      throw( Bp2FitsErr( string( "Illegal character in FITS card keyword `" ) + keyword + "'"  ) );

    }
  }

  return fitskeyword;

}

std::string makeKeyValue( 
  const string& keyword,
  const string& value, 
  size_t& offset,
  const size_t maxsize,
  const bool contchar,
  const bool _verbose,
  const bool _longstrn
  ) {
    if ( _verbose && ! _longstrn && maxsize < value.size( ) )  {
        TraceFct tf( "makeKeyValue" );
        TraceFct::message( string( "WARNING: Truncating BPipe hdrf value for `" ) + keyword + "'\n" );
    }

  string subvalue = value.substr( offset );
  size_t length;

  if ( maxsize < subvalue.size( ) && _longstrn && contchar ) {
    length = maxsize-1;
    subvalue[maxsize-1] = '&';

  } else {
    length = maxsize;

  }

  offset += length;
  return std::string( subvalue, 0, maxsize );
}

